/* 
  multi-thread event for MPI 
  M.TSUJI 2011.12
 */

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include "omrpc_defs.h"
#include "omrpc_io.h"
#include "omrpc_rpc.h"
#include "omrpc_mpi.h"
#include "omrpc_mpi_io.h"

static int to_fd,from_fd;
static int modified_flag;

fd_set rfds;

int omrpc_handler_running;

void omrpcm_handle_rex_end(omrpc_mpi_handle_t *mp)
{
  char ack;
  omrpc_request_t *req;
  omrpc_rpc_t     *rp;
  rp  = (omrpc_rpc_t *)mp->hint;
  req = rp->req;

  //omprcm_debug_print_mp();

  if(omrpc_debug_flag)
    omrpc_prf("handle request mpi=%p req=%p %s pid=%p\n",
              mp,req,rp->module->name,rp->pid);
  switch(req->kind){
  case OMRPC_CALL:
    omrpcm_call_recv_args(mp,req->stub->stub_info,req->args);
    break;
  default:
    omrpc_fatal("omrpcm_handle_rex_end: bad state kind");
  }

  /* release rpc object */
  rp->req = NULL;     /* unbound */

  if(req->kind == OMRPC_CALL && !rp->is_handle){
    omrpc_put_idle_rpc(rp);
  } 
  /* notify waiting thread */

  //pthread_mutex_lock(&req->lock);
  req->done_flag = TRUE;
  //pthread_cond_signal(&req->cond);
  //pthread_mutex_unlock(&req->lock);

  if(omrpc_debug_flag)
    omrpc_prf("handle request mpi=%p req=%p .......... end\n",
           mp,req);

  return;
} /* omrpcm_handle_rex_end */

/* hander thread main (handle the ending of remote programs?) */
static void *omrpcm_handler_main(void *dummy)
{
    int                r, flag = 0;
    char               ack;
    MPI_Status         status;
    omrpc_mpi_handle_t *mpi_ptr;
    omrpc_rpc_t        *rp;

    while(mpi_handle_head==NULL){
        sched_yield();
    }

 again:
    
    if(!mpi_handle_head) return;

    omrpcm_lockall(1);

    mpi_ptr=mpi_handle_head;
    while(mpi_ptr){
      rp = (omrpc_rpc_t *)mpi_ptr->hint;
      if(rp && rp->nprocs>0){
        flag=0;
        MPI_Iprobe(0, OMRPC_END_STUB_TAG, mpi_ptr->comm, &flag, &status);
        if(flag){
          omrpcm_handle_rex_end(mpi_ptr);
          break;
        }
      }
      mpi_ptr=mpi_ptr->next;
    }

    omrpcm_unlockall(1);

    sched_yield();

    goto again; /* loop forever */

    return;
} /* omrpcm_handler_main */

void omrpcm_io_event_init()
{
    pthread_attr_t attr;
    int fds[2];
    int r;
    omrpc_io_port_t *pp;

    if(pipe(fds) != 0){
        perror("pipe");
        omrpc_fatal("cannot create pipe");
    }
    // communication pipe between caller to handler 
    to_fd = fds[1];
    from_fd = fds[0];

    //pthread_mutex_init(&io_lock,NULL);

    // create thread 
    pthread_attr_init(&attr);
    r = pthread_create(&mrpc_tid, &attr, omrpcm_handler_main, NULL);
    
    if(r != 0){
        omrpc_fatal("cannot create handler thread");
    }
    omrpc_handler_running = TRUE;
    
    if(omrpc_debug_flag) omrpc_prf("omrpc_io_event_init end ...\n");
} /* omrpcm_io_event_init */


